#[=======================================================================[.rst:
WriteExportDependencies
-----------------------

Declare the helper function to write an export dependencies file.
By default CMake export target without dependencies information.
E.g. your exported target would have a list of other imported libraries,
but no instructions on how to find them. It makes developer's life little bit
harder, cuz you have to find dependencies of your dependencies by yourself.
This module provide a helper function aimed to address that kind of issue.

Functions
^^^^^^^^^

#]=======================================================================]

#=============================================================================
# Copyright 2014-2017 by Alex Turbov <i.zaufi@gmail.com>
#
# Distributed under the OSI-approved BSD License (the "License");
# see accompanying file LICENSE for details.
#
# This software is distributed WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the License for more information.
#=============================================================================
# (To distribute this file outside of this repository, substitute the full
#  License text for the above reference.)

#BEGIN Internals
set(_wed_BASE_DIR "${CMAKE_CURRENT_LIST_DIR}")
#END Internals

function(_ecm_collect_deps TARGET)
    set(_options RECURSIVE)
    set(_one_value_args RESULT_VARIABLE)
    set(_multi_value_args)
    cmake_parse_arguments(_ecm_collect_deps "${_options}" "${_one_value_args}" "${_multi_value_args}" ${ARGN})

    ecm_debug("Collect deps for `${TARGET}`")
    ecm_debug_indent()

    get_target_property(_type ${TARGET} TYPE)
    if(_type STREQUAL "INTERFACE_LIBRARY")
        set(_libs)
        ecm_debug("Skip checking `LINK_LIBRARIES` for interface library `${TARGET}`")
    else()
        get_target_property(_libs ${TARGET} LINK_LIBRARIES)
        if(NOT _libs)
            ecm_debug("Skip checking `LINK_LIBRARIES` for `${TARGET}` cuz empty")
            set(_libs)
        else()
            ecm_debug("Have `LINK_LIBRARIES` for `${TARGET}`: ${_libs}")
        endif()
    endif()

    get_property(_have_ilibs TARGET ${TARGET} PROPERTY INTERFACE_LINK_LIBRARIES SET)
    if(_have_ilibs)
        get_target_property(_ilibs ${TARGET} INTERFACE_LINK_LIBRARIES)
        if(NOT _ilibs)
            set(_ilibs)
            ecm_debug("Skip checking `INTERFACE_LINK_LIBRARIES` for `${TARGET}` cuz empty")
        else()
            ecm_debug("Have `INTERFACE_LINK_LIBRARIES` for `${TARGET}`: ${_ilibs}")
        endif()
    else()
        set(_ilibs)
        ecm_debug("Don't have `INTERFACE_LINK_LIBRARIES` for `${TARGET}`")
    endif()

    if(_ecm_collect_deps_RECURSIVE)

        # Get the current value of result variable (to prevent override its value)
        # TODO Do it only w/ an option `APPEND`?
        set(_result ${${_ecm_collect_deps_RESULT_VARIABLE}})

        foreach(_lib IN LISTS _libs _ilibs)
            ecm_debug("Consider `${_lib}`")
            if(_lib IN_LIST _result)
                ecm_debug("skip seen lib `${_lib}`")
                continue()

            elseif(_lib MATCHES "^\\$<LINK_ONLY:((.*)::(.*))>$")
                ecm_debug("Got LINK-ONLY lib: ${CMAKE_MATCH_1}")
                _ecm_collect_deps(${CMAKE_MATCH_1} RESULT_VARIABLE _result RECURSIVE)

            # TODO There is a better (reliable) way!
            elseif(_lib MATCHES "^(.*)::(.*)$")
                ecm_debug("Got imported target: ${CMAKE_MATCH_1}")
                _ecm_collect_deps(${_lib} RESULT_VARIABLE _result RECURSIVE)

            else()
                list(APPEND _result "${_lib}")
                ecm_debug("just lib `${_lib}`")
            endif()
        endforeach()

        # Add self to the list to avoid further attempts to collect already present libs
        list(APPEND _result "${TARGET}")

    else()
        list(APPEND _result ${_libs} ${_ilibs})
        if(_result)
            list(REMOVE_DUPLICATES _result)
        endif()
    endif()

    set(${_ecm_collect_deps_RESULT_VARIABLE} ${_result} PARENT_SCOPE)

    ecm_debug_unindent()
    ecm_debug("Collected dependencies: `${_result}` for ${TARGET}")
endfunction()

function(_ecm_try_get_package_name_from_target TARGET RESULT_VARIABLE)
    if(TARGET ${TARGET} AND "${TARGET}" MATCHES "^(.*)::(.*)$")
        set(_vendor "${CMAKE_MATCH_1}")
        string(TOUPPER "${_vendor}" _vendor_up)
        string(MAKE_C_IDENTIFIER "${_vendor_up}" _vendor_up)

        set(_comp "${CMAKE_MATCH_2}")
        string(TOUPPER "${_comp}" _comp_up)
        string(MAKE_C_IDENTIFIER "${_comp_up}" _comp_up)

        set(_module_id "${_vendor_up}_${_comp_up}")
        get_property(_has_package_name TARGET ${TARGET} PROPERTY INTERFACE_${_module_id}_PACKAGE_NAME SET)
        if(_has_package_name)
            get_property(_package_name TARGET ${TARGET} PROPERTY INTERFACE_${_module_id}_PACKAGE_NAME)
            # Set the result variable
            set(${RESULT_VARIABLE} ${_package_name} PARENT_SCOPE)
        else()
            message(
                SEND_ERROR
                "Given target `${TARGET}` has no `INTERFACE_${_module_id}_PACKAGE_NAME` property set! Code review required!"
              )
            return()
        endif()

    else()
        message(SEND_ERROR "Given target `${TARGET}` has unexpected name format (or not a target)! Code review required!")
    endif()
endfunction()

function(_wed_split_targets_int_ext)
    set(_options)
    set(_one_value_args INTERNAL_TARGETS_RESULT_VARIABLE EXTERNAL_TARGETS_RESULT_VARIABLE NON_TARGETS_RESULT_VARIABLE)
    set(_multi_value_args TARGETS)
    cmake_parse_arguments(_wed "${_options}" "${_one_value_args}" "${_multi_value_args}" ${ARGN})

    if(NOT _wed_TARGETS)
        message(SEND_ERROR "OOPS! `TARGETS` required")
    endif()

    ecm_debug_indent()

    set(_external_deps)
    set(_internal_deps)
    set(_non_target_deps)

    foreach(_d ${_wed_TARGETS})
        ecm_debug_indent()
        ecm_debug("consider: ${_d}")

        get_target_property(_is_d_imported ${_d} IMPORTED)

        if(TARGET ${_d} AND _is_d_imported)
            ecm_debug("found imported library `${_d}`")
            list(APPEND _external_deps "${_d}")


        elseif(TARGET ${_d})
            ecm_debug("found local target `${_d}`")
            list(APPEND _internal_deps "${_d}")

        else()
            ecm_debug("found non-target `${_d}`")
            list(APPEND _non_target_deps "${_d}")

        endif()

        ecm_debug_unindent()
    endforeach()

    ecm_debug_unindent()

    if(_wed_INTERNAL_TARGETS_RESULT_VARIABLE)
        set(${_wed_INTERNAL_TARGETS_RESULT_VARIABLE} ${_internal_deps} PARENT_SCOPE)
    endif()

    if(_wed_EXTERNAL_TARGETS_RESULT_VARIABLE)
        set(${_wed_EXTERNAL_TARGETS_RESULT_VARIABLE} ${_external_deps} PARENT_SCOPE)
    endif()

    if(_wed_NON_TARGETS_RESULT_VARIABLE)
        set(${_wed_NON_TARGETS_RESULT_VARIABLE} ${_non_target_deps} PARENT_SCOPE)
    endif()

endfunction()


#[=======================================================================[.rst:
.. cmake:command:: write_export_dependencies

When CMake *export* targets from some project, it writes a couple of generated
files, so depended projects may *import* them and use just like any other
target just mention a name in :cmake:command:`target_link_libraries` call,
prefixed with a vendor's namespace. Unfortunately, on export it mention
other imported targets "as is" -- i.e. if current project has library ``foo``,
which depends on ``Vendor::bar``, the result file would contain the imported target
``Vendor::foo`` which links to ``Vendor::bar`` and no clue how to find the latter.

.. code-block:: cmake
   :caption: **Synopsys**

      write_export_dependencies(<TARGET>)

- :cmake:variable:`TARGET` name of the target to write dependencies for.

This function collect direct dependencies of the given target and form an
additional CMake file, which is going to be loaded at :cmake:command:`find_package`
time and requested package/component. This generated file responsible to
load other in-project dependencies and/or find dependency packages for
imported libraries.

.. attention::

  There is a subtle thing has used: CMake's generated exported targets loader do
  wildcard :cmake:command:`include` of available configurations for given target.
  This function must produce a file, which name must match the ``blah-*.cmake`` expression.
  It is why the generated file has ``*-dependencies.cmake`` suffix.

#]=======================================================================]
function(write_export_dependencies TARGET_NAME)
    set(options)
    set(one_value_args)
    set(multi_value_args)
    cmake_parse_arguments(_wed "${options}" "${one_value_args}" "${multi_value_args}" ${ARGN})

    # Get a list of direct dependencies
    if(TARGET ${TARGET_NAME})
        _ecm_collect_deps(${TARGET_NAME} RESULT_VARIABLE _wed_link_libraries)
    else()
        message(FATAL_ERROR "Given name `${TARGET_NAME}` is not a target")
    endif()

    ecm_debug("Split target `${TARGET_NAME}` dependencies")
    _wed_split_targets_int_ext(
        TARGETS ${_wed_link_libraries}
        INTERNAL_TARGETS_RESULT_VARIABLE _internal_deps
        EXTERNAL_TARGETS_RESULT_VARIABLE _external_deps
        NON_TARGETS_RESULT_VARIABLE
      )

    # Ok, iterate over imported targets and find package names for further `find_package`
    foreach(_d IN LISTS _external_deps)
        if(NOT _d MATCHES "^(.*)::(.*)$")
            message(SEND_ERROR "OOPS! Imported library ${_d} has unexpected name!?")
        endif()

        set(_vendor "${CMAKE_MATCH_1}")
        string(TOUPPER "${_vendor}" _vendor_up)
        string(MAKE_C_IDENTIFIER "${_vendor_up}" _vendor_up)

        set(_comp "${CMAKE_MATCH_2}")
        string(TOUPPER "${_comp}" _comp_up)
        string(MAKE_C_IDENTIFIER "${_comp_up}" _comp_up)

        # Handle `this project` packages first
        if(_vendor STREQUAL "Vendor")

            _ecm_try_get_package_name_from_target(${_d} _pkg_name)

        else()
            # Handle other "vendors":
            # - an exact package name to find is not known (guessed the same as a vendor's name)
            set(_pkg_name ${_vendor})
        endif()

        string(TOUPPER "${_pkg_name}" _pkg_name_up)
        string(MAKE_C_IDENTIFIER "${_pkg_name_up}" _pkg_name_up)

        if(_pkg_name)
            ecm_debug("Append `${_pkg_name}` to list of required packages w/ vendor `${_vendor}` and component `${_comp}`")
            list(APPEND _packages_to_find ${_pkg_name})
            list(APPEND _wed_${_pkg_name_up}_COMPONENTS ${_comp})
            set(_wed_${_pkg_name_up}_VENDOR ${_vendor})

        else()
            # ATTENTION If you see this, review `*-config.cmake.in` of the corresponding package
            message(
                SEND_ERROR
                "Target `${_d}` do not have package name property! CMake config file code review required!"
                )
        endif()

    endforeach()

    set(_unaliased_internal_deps)
    foreach(_d IN LISTS _internal_deps)
        # Local target also could be aliased
        if(_d MATCHES "^(.*)::(.*)$")
            set(_d "${CMAKE_MATCH_2}")
        endif()
        list(APPEND _unaliased_internal_deps "${_d}")
    endforeach()
    set(_wed_INTERNAL_DEPS ${_unaliased_internal_deps})
    unset(_unaliased_internal_deps)

    # Do we really have smth to write?
    if(NOT (_packages_to_find OR _internal_deps))
        ecm_debug("The given target `${TARGET_NAME}` have no dependencies! Why to call me then?")
        return()
    endif()

    string(RANDOM ALPHABET "qwertyuiopasdfghjklzxcvbnm" _wed_var_rnd)

    # Ok, render the internal deps loader if needed
    if(_internal_deps)
        file(READ "${_wed_BASE_DIR}/export-dependencies-internal.cmake.in" _internal_tpl)
        string(CONFIGURE "${_internal_tpl}" _wed_FIND_INTERNAL_DEPS @ONLY)
    endif()

    # Now the most hardest part ;-0 -- external dependencies
    if(_packages_to_find)
        list(REMOVE_DUPLICATES _packages_to_find)

        set(_wed_SETUP_VARIABLES "\n# Packages, versions and components to find")
        set(_wed_DESTROY_VARIABLES "\n# ATTENTION It is important not to leave any temporary variables after this script")

        string(
            APPEND _wed_SETUP_VARIABLES
            "\nset(_${_wed_var_rnd}_PACKAGES ${_packages_to_find})"
          )
        string(
            APPEND _wed_DESTROY_VARIABLES
            "\nunset(_${_wed_var_rnd}_PACKAGES)"
          )

        foreach(_pkg_name IN LISTS _packages_to_find)
            string(TOUPPER "${_pkg_name}" _pkg_name_up)
            string(MAKE_C_IDENTIFIER "${_pkg_name_up}" _pkg_name_up)

            string(APPEND _wed_SETUP_VARIABLES "\nset(_${_wed_var_rnd}_${_pkg_name_up}_VENDOR ${_wed_${_pkg_name_up}_VENDOR})")
            string(APPEND _wed_DESTROY_VARIABLES "\nunset(_${_wed_var_rnd}_${_pkg_name_up}_VENDOR)")

            list(REMOVE_DUPLICATES _wed_${_pkg_name_up}_COMPONENTS)
            string(APPEND _wed_SETUP_VARIABLES "\nset(_${_wed_var_rnd}_${_pkg_name_up}_COMPONENTS ${_wed_${_pkg_name_up}_COMPONENTS})")
            string(APPEND _wed_DESTROY_VARIABLES "\nunset(_${_wed_var_rnd}_${_pkg_name_up}_COMPONENTS)")

            if(DEFINED ${_pkg_name_up}_VERSION)
                string(APPEND _wed_SETUP_VARIABLES "\nset(_${_wed_var_rnd}_${_pkg_name_up}_VERSION ${${_pkg_name_up}_VERSION})")
                string(APPEND _wed_DESTROY_VARIABLES "\nunset(_${_wed_var_rnd}_${_pkg_name_up}_VERSION)")

            elseif(DEFINED ${_pkg_name}_VERSION)
                string(APPEND _wed_SETUP_VARIABLES "\nset(_${_wed_var_rnd}_${_pkg_name_up}_VERSION ${${_pkg_name}_VERSION})")
                string(APPEND _wed_DESTROY_VARIABLES "\nunset(_${_wed_var_rnd}_${_pkg_name_up}_VERSION)")
            endif()

        endforeach()

        file(READ "${_wed_BASE_DIR}/export-dependencies-external.cmake.in" _external_tpl)
        string(CONFIGURE "${_external_tpl}" _wed_FIND_EXTERNAL_DEPS @ONLY)
    endif()

    # And finally render the main template
    ecm_debug("Write to `${CMAKE_CURRENT_BINARY_DIR}/blah-${TARGET_NAME}-targets-dependencies.cmake`")
    configure_file(
        "${_wed_BASE_DIR}/export-dependencies-main.cmake.in"
        "${CMAKE_CURRENT_BINARY_DIR}/blah-${TARGET_NAME}-targets-dependencies.cmake"
        @ONLY
      )
endfunction()

# X-Chewy-RepoBase: https://raw.githubusercontent.com/mutanabbi/chewy-cmake-rep/master/
# X-Chewy-Path: WriteExportDependencies.cmake
# X-Chewy-Version: 4.0
# X-Chewy-Description: Write an export dependencies file
# X-Chewy-AddonFile: export-dependencies-main.cmake.in
# X-Chewy-AddonFile: export-dependencies-internal.cmake.in
# X-Chewy-AddonFile: export-dependencies-external.cmake.in
